/*
 * QQQ - Low-code Application Framework for Engineers.
 * Copyright (C) 2021-2022.  Kingsrook, LLC
 * 651 N Broad St Ste 205 # 6917 | Middletown DE 19709 | United States
 * contact@kingsrook.com
 * https://github.com/Kingsrook/
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package com.kingsrook.qqq.backend.core.model.metadata.processes;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.kingsrook.qqq.backend.core.model.metadata.fields.QFieldMetaData;


/*******************************************************************************
 ** Meta-Data to define a front-end step in a process in a QQQ instance (e.g.,
 ** a screen presented to a user).
 **
 *******************************************************************************/
public class QFrontendStepMetaData extends QStepMetaData
{
   private List<QFrontendComponentMetaData> components;
   private List<QFieldMetaData>             formFields;
   private List<QFieldMetaData>             viewFields;
   private List<QFieldMetaData>             recordListFields;
   private Map<String, QFieldMetaData>      formFieldMap;



   /*******************************************************************************
    **
    *******************************************************************************/
   public QFrontendStepMetaData()
   {
      setStepType("frontend");
   }



   /*******************************************************************************
    ** Getter for components
    **
    *******************************************************************************/
   public List<QFrontendComponentMetaData> getComponents()
   {
      return components;
   }



   /*******************************************************************************
    ** Setter for components
    **
    *******************************************************************************/
   public void setComponents(List<QFrontendComponentMetaData> components)
   {
      this.components = components;
   }



   /*******************************************************************************
    ** Fluent setter for adding 1 component
    **
    *******************************************************************************/
   public QFrontendStepMetaData withComponent(QFrontendComponentMetaData component)
   {
      if(this.components == null)
      {
         this.components = new ArrayList<>();
      }
      this.components.add(component);
      return (this);
   }



   /*******************************************************************************
    ** Fluent setter for components
    **
    *******************************************************************************/
   public QFrontendStepMetaData withComponents(List<QFrontendComponentMetaData> components)
   {
      this.components = components;
      return (this);
   }



   /*******************************************************************************
    ** Getter for a single formFields by its name
    **
    *******************************************************************************/
   public QFieldMetaData getFormField(String fieldName)
   {
      if(formFieldMap != null && formFieldMap.containsKey(fieldName))
      {
         return (formFieldMap.get(fieldName));
      }

      return (null);
   }



   /*******************************************************************************
    ** Getter for formFields
    **
    *******************************************************************************/
   public List<QFieldMetaData> getFormFields()
   {
      return formFields;
   }



   /*******************************************************************************
    ** adder for formFields
    **
    *******************************************************************************/
   public void addFormField(QFieldMetaData fieldMetaData)
   {
      if(fieldMetaData != null)
      {
         if(formFieldMap == null)
         {
            formFieldMap = new HashMap<>();
         }
         if(formFields == null)
         {
            formFields = new ArrayList<>();
         }

         formFieldMap.put(fieldMetaData.getName(), fieldMetaData);
         formFields.add(fieldMetaData);
      }
   }



   /*******************************************************************************
    ** Setter for formFields
    **
    *******************************************************************************/
   public void setFormFields(List<QFieldMetaData> formFields)
   {
      if(formFields != null)
      {
         for(QFieldMetaData fieldMetaData : formFields)
         {
            addFormField(fieldMetaData);
         }
      }
   }



   /*******************************************************************************
    ** fluent setter to add a single form field
    **
    *******************************************************************************/
   public QFrontendStepMetaData withFormField(QFieldMetaData formField)
   {
      addFormField(formField);
      return (this);
   }



   /*******************************************************************************
    ** fluent setter for formFields
    **
    *******************************************************************************/
   public QFrontendStepMetaData withFormFields(List<QFieldMetaData> formFields)
   {
      this.setFormFields(formFields);
      return (this);
   }



   /*******************************************************************************
    ** Getter for viewFields
    **
    *******************************************************************************/
   public List<QFieldMetaData> getViewFields()
   {
      return viewFields;
   }



   /*******************************************************************************
    ** Setter for viewFields
    **
    *******************************************************************************/
   public void setViewFields(List<QFieldMetaData> viewFields)
   {
      this.viewFields = viewFields;
   }



   /*******************************************************************************
    ** fluent setter to add a single view field
    **
    *******************************************************************************/
   public QFrontendStepMetaData withViewField(QFieldMetaData viewField)
   {
      if(this.viewFields == null)
      {
         this.viewFields = new ArrayList<>();
      }
      this.viewFields.add(viewField);
      return (this);
   }



   /*******************************************************************************
    ** fluent setter for viewFields
    **
    *******************************************************************************/
   public QFrontendStepMetaData withViewFields(List<QFieldMetaData> viewFields)
   {
      this.viewFields = viewFields;
      return (this);
   }



   /*******************************************************************************
    ** Getter for recordListFields
    **
    *******************************************************************************/
   public List<QFieldMetaData> getRecordListFields()
   {
      return recordListFields;
   }



   /*******************************************************************************
    ** Setter for recordListFields
    **
    *******************************************************************************/
   public void setRecordListFields(List<QFieldMetaData> recordListFields)
   {
      this.recordListFields = recordListFields;
   }



   /*******************************************************************************
    ** fluent setter to add a single recordList field
    **
    *******************************************************************************/
   public QFrontendStepMetaData withRecordListField(QFieldMetaData recordListField)
   {
      if(this.recordListFields == null)
      {
         this.recordListFields = new ArrayList<>();
      }
      this.recordListFields.add(recordListField);
      return (this);
   }



   /*******************************************************************************
    ** fluent setter for recordListFields
    **
    *******************************************************************************/
   public QFrontendStepMetaData withRecordListFields(List<QFieldMetaData> recordListFields)
   {
      this.recordListFields = recordListFields;
      return (this);
   }



   /*******************************************************************************
    ** fluent setter for name
    **
    *******************************************************************************/
   @Override
   public QFrontendStepMetaData withName(String name)
   {
      setName(name);
      return (this);
   }



   /*******************************************************************************
    ** fluent setter for label
    **
    *******************************************************************************/
   @Override
   public QFrontendStepMetaData withLabel(String label)
   {
      setLabel(label);
      return (this);
   }



   /*******************************************************************************
    ** Get a list of all of the input fields used by this function
    *******************************************************************************/
   @JsonIgnore // because this is a computed property - we don't want it in our json.
   @Override
   public List<QFieldMetaData> getInputFields()
   {
      List<QFieldMetaData> rs = new ArrayList<>();

      if(formFields != null)
      {
         rs.addAll(formFields);
      }

      return (rs);
   }

}
